;;; JACAL: Symbolic Mathematics System.        -*-scheme-*-
;;; Copyright 1989, 1990, 1991, 1992, 1993, 1995 Aubrey Jaffer.

;;; The following are useful for debugging representational and other
;;; problems.

;;; (POLY:VALID? obj) returns #t if the variable ordering in obj is
;;; consistent with respect to var:>.

;;; SCM has (STACK-TRACE), which prints evaluation pending expressions.

;;; To print out an expression in the current output format, do:
;;; (write-sexp (math->sexp exp #f) *output-grammar*) (force-output)

;;; (MATH:BREAK-WITH-OBJECT MSG OBJ) prints MSG, sets the jacal
;;; variable errobj to OBJ, and breaks to the jacal prompt.

;;; (MATH:CONTINUABLE-BREAK-WITH-OBJECT MSG OBJ) prints MSG, sets the
;;; jacal variable errobj to OBJ, saves its continuation and breaks to
;;; the jacal prompt.

(define math:break-continuation-stack '())

(define (math:break-with-object msg obj)
  (display-diag msg) (newline-diag)
  (display-diag " Offending object is now value of errobj.") (newline-diag)
  (defsym 'errobj obj)
  (math:exit #f))

(define (math:continuable-break-with-object msg obj)
  (display-diag msg) (newline-diag)
  (display-diag " Offending object is now value of errobj.") (newline-diag)
  (display-diag " To continue type continue(RETURN-VALUE);") (newline-diag)
  (defsym 'errobj obj)
  (call-with-current-continuation
   (lambda (cont)
     (set! math:break-continuation-stack
	   (cons cont math:break-continuation-stack))
     (math:exit #f))))

(require 'defmacro)
(define debug:indent 0)

(define (math:printn x) (math:print x) (newline-diag))

(define (print . args)
  (define result #f)
  (for-each (lambda (x) (set! result x) (math:print x)
		    (display-diag #\ ))
	    args)
  (newline-diag)
  result)

(define (mtracef function . optname)
      (set! debug:indent 0)
      (let ((name (if (null? optname) function (car optname))))
	(lambda args
	  (cond ((and (not (null? args))
		      (eq? (car args) 'debug:untrace-object)
		      (null? (cdr args)))
		 function)
		(else
		 (do ((i debug:indent (+ -1 i))) ((zero? i)) (display #\ ))
		 (display "CALLED ") (display name) (newline)
		 (for-each math:printn args)
		 (set! debug:indent (modulo (+ 1 debug:indent) 8))
		 (let ((ans (apply function args)))
		   (set! debug:indent (modulo (+ -1 debug:indent) 8))
		   (do ((i debug:indent (+ -1 i))) ((zero? i)) (display #\ ))
		   (display "RETURNED ") (display name) (newline)
		   (math:printn ans)
		   ans))))))

(define (muntracef function)
  (set! debug:indent 0)
  (function 'debug:untrace-object))

(require 'alist)
(define mtrace:adder (alist-associator eq?))
(define mtrace:deler (alist-remover eq?))

(define *traced-procedures* '())
(define (mtrace:tracef fun sym)
  (cond ((not (procedure? fun))
	 (display "WARNING: not a procedure " (current-error-port))
	 (display sym (current-error-port))
	 (newline (current-error-port))
	 (set! *traced-procedures* (mtrace:deler *traced-procedures* sym))
	 fun)
	(else
	 (let ((p (assq sym *traced-procedures*)))
	   (cond ((and p (eq? (cdr p) fun))
		  fun)
		 (else
		  (let ((tfun (mtracef fun sym)))
		    (set! *traced-procedures*
			  (mtrace:adder *traced-procedures* sym tfun))
		    tfun)))))))

(define (mtrace:untracef fun sym)
  (let ((p (assq sym *traced-procedures*)))
    (set! *traced-procedures* (mtrace:deler *traced-procedures* sym))
    (cond ((not (procedure? fun)) fun)
	  ((not p) fun)
	  ((eq? (cdr p) fun)
	   (muntracef fun))
	  (else fun))))

;;; Macros.

(for-each
 defmacro:eval
 '(
   (defmacro trace xs
     (if (null? xs)
	 `(begin ,@(map (lambda (x) `(set! ,x (mtrace:tracef ,x ',x)))
			(map car *traced-procedures*))
		 (map car *traced-procedures*))
	 `(begin ,@(map (lambda (x) `(set! ,x (mtrace:tracef ,x ',x))) xs))))
   (defmacro untrace xs
     (if (null? xs)
	 (slib:eval
	  `(begin ,@(map (lambda (x) `(set! ,x (mtrace:untracef ,x ',x)))
			 (map car *traced-procedures*))
		  '',(map car *traced-procedures*)))
	 `(begin ,@(map (lambda (x) `(set! ,x (mtrace:untracef ,x ',x))) xs))))
   (defmacro mtrace xs
     (if (null? xs)
	 `(begin ,@(map (lambda (x) `(set! ,x (mtrace:tracef ,x ',x)))
			(map car *traced-procedures*))
		 (map car *traced-procedures*))
	 `(begin ,@(map (lambda (x) `(set! ,x (mtrace:tracef ,x ',x))) xs))))
   (defmacro muntrace xs
     (if (null? xs)
	 (slib:eval
	  `(begin ,@(map (lambda (x) `(set! ,x (mtrace:untracef ,x ',x)))
			 (map car *traced-procedures*))
		  '',(map car *traced-procedures*)))
	 `(begin ,@(map (lambda (x) `(set! ,x (mtrace:untracef ,x ',x))) xs))))
   ))

(define (debug:mtrace-all file)
  (call-with-input-file
      file
    (lambda (port)
      (define (get-first-symbol form)
	(if (pair? form)
	    (get-first-symbol (car form))
	    form))
      (require 'mtrace)
      (require 'line-i/o)
      (let loop ((fc (peek-char port)))
	(cond ((eof-object? fc))
	      ((char-whitespace? fc))
	      ((char=? #\; fc))
	      ((char=? #\( fc))
	      (else (read-line port) (loop (peek-char port)))))
      (defmacro:eval
        (cons 'mtrace
              (do ((form (read port) (read port))
                   (tlist '()
			  (if (and (pair? form)
				   (eq? (car form) 'define)
				   (or (pair? (cadr form))
				       (and (pair? (caddr form))
					    (list? (caddr form))
					    (eq? 'lambda (caaddr form)))))
			      (cons (get-first-symbol (cadr form))
				    tlist)
			      tlist)))
                  ((eof-object? form) tlist)))))))

(define mtrace-all debug:mtrace-all)

(define (debug:check proc1 proc2 . opts)
  (lambda args
    (let ((result1 (apply proc1 args))
	  (result2 (apply proc2 args)))
      (cond ((not (equal? result1 result2))
	     (newline-diag)
	     (display "WARN: ")
	     (for-each write opts)
	     (display ": results not equal?")
	     (newline-diag)
	     (display (if (null? opts) "{1}" (car opts)))
	     (for-each math:print args)
	     (newline-diag)
	     (math:print result1)
	     (newline-diag)
	     (display (if (or (null? opts) (null? (cdr opts))) "{1}"
			  (cadr opts)))
	     (math:print result2)
	     (newline-diag)))
      result1)))
