;;; "ff.scm" Polynomial factorization.		-*-scheme-*-
;;; Copyright 1994, 1995 Mike Thomas
;;; See the file "COPYING" for terms applying to this program.

;;; AUTHOR
;;;	  Mike Thomas
;;;	  46 Harold Street
;;;	  STAFFORD  QLD	 4053
;;;	  AUSTRALIA

;;;	  Email: mjt@octavia.anu.edu.au

;;;	  Phone: Intl + 61 7 356 8494

;;; SOURCES
;;; These algorithms are drawn from:
;;;    Geddes, Czapor & Labahn, 'Algorithms for Computer Algebra',
;;;	 Kluwer, 1992					(GCL)
;;;    Davenport, Siret & Tournier, 'Computer Algebra Systems and
;;;	 Algorithms for Algebraic Computation'and Knuth',
;;;	 Academic Press, 1993
;;;    Knuth, 'Seminumerical Algorithms'.

;;; DEVELOPMENT LANGUAGE
;;; SCM 4e1, 4e2 compiled with DICE v 2.07.54

;;; DEVELOPMENT SYSTEM
;;; Amiga 1200, 2Mb Chip RAM, 4Mb Fast RAM, 50MHz 68030 CPU No FPU,
;;; Workbench 3.0 (until a power supply problem stuffed it that is.)

;;; DEPENDENCIES
;;; You need R4RS Scheme, SLIB and JACAL (latest versions).
;;; ff:p+p and ff:p-p depend on the behaviour of map when the argument
;;;    lists are of different lengths.
;;; Multiargument *, +

;;; REPRESENTATIONS
;;; The modular homomorphism sym:sym maps the integers to a symmetric
;;; representation, following the numerous examples in GCL. This
;;; mapping does not seem to correspond with GCL's own definition of Z
;;; mod n using a simple remainder algorithm.

;;; The univariate polynomials expected and returned by the Scheme
;;; procedures are JACAL's list representation, not numbers by
;;; themselves.  There is no attempt to cover sparse polynomials yet.

;;; An ideal is a list of lists, each one specifying a variable and
;;; it's value eg '((y 1) (z 2) (p -1)).

;;; DESIGN PHILOSOPHY

;;; 1. Fun;

;;; 2. Correct operation for univariate and multivariate polynomials
;;;    over the rationals (Q), then maybe over algebraic number fields;

;;; 3. refine the algorithms for speed after they work correctly;

;;; 4. make the system workable even for smaller computers, based on the
;;;    crude and discriminatory assumption that people who want to
;;;    crack large problems can afford to buy or write larger systems;

;;; 5. and ideally, I want the system to handle investigation of
;;;    the stability of systems of differential equations
;;;    which requires algebraic solution of eigensystems,
;;;    and for an attempt on the Risch integration algorithm.

;;; I have therefore initially used Linear Hensel Lifting and the small prime
;;; Berlekamp Algorithm. I make no attempt yet, among other possibilities,
;;; to optimise the factorisation process dynamically by checking
;;; several primes or restarting the algorithm on testing for true factors
;;; after the Berlekamp process. Later. Perhaps.

;;; NAMING
;;;	     u: univariate polynomials, should be JACAL's univ:
;;;	    ff: polynomials over finite integer fields using the
;;;		symmetric representation
;;;	   sym: symmetric representation integers
;;;	  poly: Jacal poly types
;;;	   hen: Hensel lifting procedures

;;; ABBREVIATIONS
;;;	     Q	The Rational numbers
;;;	     Z	The integers

;;; ARGUMENT CODING CONVENTIONS
;;;	    ff: the modulus argument is the last in the argument list

;;; KNOWN BUGS

;;; -- factorq((2*z+y-x)*(y^3*z-a*x^2)*(b*z^2+y));
;;;    factorq((a*a*b*z+d)*(2*a*b*b*z+c)*((u+a)*z+1));
;;;    Hang my presently memory challenged computer during poly:sff
;;;	probably because of lack of room

;;; -- Parfrac worketh not. Wherefore? Because I trieth too hard to
;;;    be clevereth. This one can wait.

;;; -- The result is not always sorted or normalised.


;;; KNOWN MINOR PROBLEMS (FROM MY POINT OF VIEW)

;;; -- Output format sometimes gets misaligned.

;;; -- Speed is not great.  See DESIGN PHILOSOPHY above.

;;; -- Factorization over Q needs to return monic polynomials if it is
;;;    to produce complete unique unit normal factorisations. Factorisation
;;;    over Z and Q would also need to factorise numeric content in this case.

(require 'array)
(require 'common-list-functions)
(require 'modular)
(require 'hash-table)
(require 'sort)
(require 'factor)
(require 'primes)
(require 'random)
(require (in-vicinity jacal-vicinity "combin.scm"))

(define sym:invert (lambda (m a) (modular:invert (quotient m -2) a)))
(define sym:sym (lambda (m a) (modular:normalize (quotient m -2) a)))

;;;; Here are listy sort of functions added by MJT:

(define (list-head l n) (butlast l (- (length l) n)))

;;; remove duplicates of members of a list l1 from a list l
(define (remove-duplicates l l1)
  (cond ((null? l) l1)
	((member (car l) l1) (remove-duplicates (cdr l) l1))
	(else (remove-duplicates (cdr l) (cons (car l) l1)))))

;;; remove object from a list l, once only (unlike delete)
(define (delete-one object l)
  (cond ((null? l) '())
	((eq? object (car l)) (cdr l))
	(else (cons (car l) (delete-one object (cdr l))))))

;;; delete all occurrences of a from b non-destructively
(define (del a b acc)
  (cond ((null? a) b)
	((null? b) acc)
	((equal? a (car b))
	  (del a (cdr b) acc))
	(else (del a (cdr b) (append acc (list (car b)))))))

;;;; Here are simple scalar sorts of functions which MJT added.

(define (sign a)
  (if (negative? a) -1 1))

;;; floored integer logarithm of x base b
(define (filog b x)
  (letrec ((log1 (lambda (b x acc)
		   (if (>= x b) (log1 b (quotient x b) (+ acc 1)) acc))))
    (log1 b x 0)))

;;; (jaffer) printer for diagnostic information
(define (ff:print . args)
  (define result #f)
  (for-each (lambda (x) (set! result x) (display-diag x) (display #\ )) args)
  (newline-diag)
  result)

;;; Idea copied from SLIB on Aubrey Jaffer's suggestion
(define (poly:sort-factors fs)
  (sort! fs poly:factor<))

(define (lnumber? x)
  (cond ((number? x) #t)
	((list? x)
	  (and (= 1 (length x)) (number? (car x))))
	(else #f)))

(define (poly:factor< x y)
  (cond ((and (number? x) (number? y)) (< x y))
	((and (lnumber? x) (lnumber? y)) (< (car x) (car y)))
	((lnumber? x) #t)
	((lnumber? y) #f)
	((null? x) #t)
	((null? y) #f)
	((and (symbol? x) (symbol? y)) (string<? (symbol->string x)
						 (symbol->string y)))
	((vector? (car x))
	  (if (string<=? (vector-ref (car x) 1) (vector-ref (car y) 1))
	      (poly:factor< (cdr x) (cdr y)) #f))
	((> (length x) (length y)) #f)
	((< (length x) (length y)) #t)
	((and (list? x)
	      (list? y))
	   (cond
	     ((poly:factor< (car (last-pair x)) (car (last-pair y))) #t)
	     ((poly:factor< (car (last-pair y)) (car (last-pair x))) #f)
	     (else (poly:factor< (butlast x 1) (butlast y 1)))))
	((list? x)
	     (poly:factor< (but-last x 1) y))
	((list? y)
	     (poly:factor< x (but-last y 1)))
	(else
	     (slib:error "poly:factor<: unknown type" x y))))

;;; Memoize an n argument procedure fn from Norvig 'artificial
;;; Intelligence Programming'
;;; Examples
;;; (define fib
;;;   (memon (lambda (n)
;;;	 (if (<= n 1) 1
;;;	    (+ (fib (- n 1)) (fib (- n 2)))))))

;;; (define sum
;;;   (memon (lambda (x y)
;;;	 (+ x y))))

(define (memon fn)
  (let ((table (make-hash-table 100))
	(gethash (hash-inquirer equal?))
	(puthash (hash-associator equal?)))
    (lambda (. x)
      (let ((val (gethash table x)))
	(if val
	    val
	    (let ((fx (apply fn x)))
	      (puthash table x fx)
	      fx))))))

;================================================================

;;; find the bound of Z mod o, where o must be odd
(define (sym:bound o)
  (quotient (- o 1) 2))

(define (ff:degree n p)
  (- (length (ff:unorm n p)) 2))

(define (u:degree p)
  (- (length p) 2))

(define (ff:leading n p)
  (sym:sym n (car (last-pair p))))

(define (u:leading p)
  (car (last-pair p)))

(define u:var car)

(define (u:zero? p)
  (and (= (length p) 2) (zero? (cadr p))))

(define (u:one? p)
  (and (= (length p) 2) (= (cadr p) 1)))

(define (u:number? p)
  (and (= (length p) 2) (number? (cadr p))))

(define (poly:poly? p)
  (cond
    ((not (list?  p)) #f)
    ((not (poly:var? (car p))) #f)
    (else #t)))

;;; Checks if a polynomial is univariate

(define (poly:univariate? p)
  (cond
    ((not (poly:poly? p)) #f)
    ((not (= 1 (length (poly:vars p)))) #f)
    (else #t)))

(define (poly:multivariate? p)
  (cond
    ((not (poly:poly? p)) #f)
    ((= (length (poly:vars p)) 1) #f)
    (else #t)))

;;; convert a zero degree univariate poly to a number
(define (poly->number p)
  (if (u:number? p)
      (cadr p)
      p))

;;; Return a polynomial in variable v if given a number, otherwise the
;;; polynomial.	 Assumes "otherwise" is a polynomial
(define (number->poly a v)
  (if (number? a)
      (list v a)
      a))

(define *primes*
  '(3 5 7 11 13 17 19 23 29 31 37 41 43 47 53 59 61
      67 71 73 79 83 89 97 101 103 107 131 137 149))

;;; Extended Euclidean algorithm returns the list (g s t),
;;; where g is the gcd of a and b, and g = s*a + t*b.
;;; GCL algorithm 2.2
;;; for univariate poly's over Z mod n:
(define (ff:eea-gcd n a b)
  (letrec ((feg (lambda (c c1 c2 d d1 d2)
		  (if (u:zero? d)
		      (list c c1 c2)
		      (let* ((qr (ff:p/p->qr n c d))
			     (q	 (car qr))
			     (r1 (ff:p-p n c1 (ff:p*p n q d1)))
			     (r2 (ff:p-p n c2 (ff:p*p n q d2))))
			(feg d d1 d2 (cadr qr) r1 r2))))))
    (let* ((var (u:var a))
	   (aone (list var 1))
	   (azero (list var 0))
	   (result (feg (ff:make-monic n a) aone azero
			(ff:make-monic n b) azero aone))
	   (c (car result)))
      (list (ff:make-monic n c)
	    (ff:p/c n (cadr result) (* (ff:leading n a) (ff:leading n c)))
	    (ff:p/c n (caddr result)(* (ff:leading n b) (ff:leading n c)))))))

;;; Standard Euclidean algorithm for gcd over Z mod n
(define (ff:euclid-gcd n a b)
  (let ((r 0))
    (let loop ((c (u:primz (ff:unorm n a)))
	       (d (u:primz (ff:unorm n b))))
      (cond ((not (u:zero? d))
	     (set! r (cadr (ff:p/p->qr n c d)))
	     (loop d (u:primz r)))
	    (else
	     (ff:make-monic n (ff:p*c n c (gcd (u:contz a) (u:contz b)))))))))

(define (ff:monic? n p)
  (if (not (u:number? p))
      (= (ff:leading n p) 1)
      #f))

(define (ff:make-monic n p)
  (if (not (ff:monic? n p))
      (ff:p*c n p (sym:invert n (ff:leading n p)))
      p))

;;; Numeric content and primitive part of a univariate polynomial p
;;; u(p) returns the sign of the leading coefficient of p
(define (u:unitz p)
  (sign (u:leading p)))

(define (u:contz p)
  (apply gcd (cdr p)))

(define (u:primz p)
  (if (u:number? p)
      (list (u:var p) (abs (cadr p)))
      (let ((c (* (u:unitz p) (u:contz p))))
	(cons (u:var p)
	      (map (lambda (x) (quotient x c)) (cdr p))))))

;;; Polynomial sign, content and primitive part of a multivariate polynomial
;;; p, with respect to var.
;;; poly:unitz returns the sign of the leading coefficient of the polynomial p.
(define (poly:unitz p var)
  (sign (leading-number (leading-coeff p var))))

(define (poly:contz p var)
  (let loop ((deg (- (poly:degree p var) 1))
	     (cgcd (leading-coeff p var)))
    (if (>= deg 0)
      (loop (- deg 1) (let ((coeft (poly:coeff p var deg)))
;;;(ff:print "deg = " deg " cgcd = " cgcd " coeft = " coeft)
			(if (number? coeft)
			  (if (zero? coeft)
			    cgcd
			    (poly:gcd cgcd coeft))
			  (poly:gcd cgcd coeft))))
      (signcan cgcd))))

(define (poly:primz p var)
  (if (number? p)
      (abs p)
      (signcan (poly:/ p (poly:contz p var)))))

;;; remove leading zeros of a poly with already modularised coefficients
(define (u:lose-lzeros p)
  (let* ((l  (length p))
	 (nl (do ((i  (- l 1) (- i 1)))
		 ((or (not (zero? (list-ref p i))) (< i 2)) (+ i 1)))))
    (if (not (= nl l))
	(reverse (list-tail (reverse p) (- l nl)))
	p)))

;;; reduce univariate p mod n and remove any consequential leading
;;; zero coefficients
(define (ff:unorm n p)
  (u:lose-lzeros (cons (u:var p) (map (lambda (x) (sym:sym n x)) (cdr p)))))

;;; Same but multivariate, copied from Jaffer's poly-mod-number in poly.scm
;;; and using sym:sym instead of modulo.
(define (ff:mnorm modulus poly)
  (if (number? poly)
    (sym:sym modulus poly)
    (let ((ltail (map-no-end-0s (lambda (x) (ff:mnorm modulus x))
			   (cdr poly))))
      (if (null? ltail) 0 (cons (car poly) ltail)))))

;;; Go through a polynomial and convert each zero degree polynomial coefficient
;;; to a number.
(define (poly:0-degree-norm p)
;;;(ff:print p)
  (cond
    ((number? p) p)
    ((= (length p) 2) (poly:0-degree-norm (cadr p)))
    (else
      (let ((end-bit (map-no-end-0s poly:0-degree-norm (cdr p))))
	(if (null? end-bit) 0
	  (cons (car p) end-bit))))))

;;; Add two numerical coefficient univariate polynomials mod n.
;;; Returns normalised sum.
(define (ff:p+p n a b)
  (let ((la (length a))
	(lb (length b)))
    (if (equal? (u:var a) (u:var b))
	(u:lose-lzeros
	 (cons (u:var a)
	       (append
		(map (lambda (x y) (sym:sym n (+ x y)))
		     (cdr a) (cdr b))
		(if (< la lb)
		    (map (lambda (x) (sym:sym n x)) (list-tail b la))
		    (map (lambda (x) (sym:sym n x)) (list-tail a lb))))))
	(slib:error 'ff:p+p "arguments have different variables."
		    (u:var a)
		    (u:var b)))))

(define (ff:p-p n a b)
  (let ((la (length a))
	(lb (length b)))
    (if (equal? (u:var a) (u:var b))
	(u:lose-lzeros
	 (cons (u:var a)
	       (append
		(map (lambda (x y) (sym:sym n (- x y)))
		     (cdr a) (cdr b))
		(if (< la lb)
		    (map (lambda (x) (sym:sym n (- 0 x))) (list-tail b la))
		    (map (lambda (x) (sym:sym n x)) (list-tail a lb))))))
	(slib:error 'ff:p-p "arguments have different variables."
		    (u:var a)
		    (u:var b)))))

(define (ff:p*c n a c)
  (if (number? c)
      (u:lose-lzeros
       (cons (u:var a)
	     (map (lambda (x) (sym:sym n (* x c)))
		  (cdr a))))
      (slib:error 'ff:p*c "constant should be a number." c)))

(define (ff:p*p-mjt n a b)
  (ff:unorm n (number->poly (poly:* a b) (u:var a))))

;(define ff:p*p
;  (debug:check ff:p*p-mjt
;	       (lambda (modulus x y)
;		 (cond ((every poly:0? (cdr y)) y)
;		       (else (mpoly:* (symmetric:modulus modulus) x y))))
;	       'ff:p*p-mjt 'mpoly:*)
;)

(define ff:p*p
  (lambda (modulus x y)
    (cond ((every poly:0? (cdr y)) y)
	  (else (mpoly:* (symmetric:modulus modulus) x y)))))

(define (u:p*c a c)
  (let* ((p (list->vector a))
	 (lp (vector-length p)))
    (if (number? c)
	(do ((i 1 (+ i 1)))
	    ((>= i lp) (vector->list p))
	  (vector-set! p i (* (vector-ref p i) c)))
	(slib:error 'ff:p*c "constant should be a number." c))))

(define (ff:p/c n a c)
  (if (number? c)
      (let ((ci (sym:invert n c)))
	(u:lose-lzeros
	 (cons (u:var a)
	       (map (lambda (x) (sym:sym n (* x ci)))
		    (cdr a)))))
      (slib:error 'ff:p/c "constant should be a number." c)))

;;; Uses quotient not /
(define (u:p/c a c)
  (if (number? c)
      (cons (u:var a) (map (lambda (x) (quotient x c)) (cdr a)))
      (error "In u:p/c ; The constant should be a number." c)))

;;; Promote the degree of a polynomial a to n
;;; assumes a is normalised.
(define (u:higher-deg a n)
  (let ((dp (- (length a) 2)))
    (cond
     ((> n dp)
      (cons
       (u:var a)
       (append (vector->list (make-vector (- n dp) 0)) (cdr a))))
     ((= n dp) a)
     (else
      (slib:error 'u:higher-deg "polynomial has higher degree than " n)))))

;;; Returns the quotient and remainder of a and b over Z mod p, Leaves
;;; all error checks and argument normalisation to calling procedure
(define (ff:p/p->qr p u v)
  (let ((m (ff:degree p u))
	(n (ff:degree p v))
	(lu (ff:leading p u))
	(lv (ff:leading p v)))
    (cond
     ((u:zero? u) (list (list (u:var v) 0) (list (u:var v) 0)))
     ((equal? u v) (list (list (u:var v) 1) (list (u:var v) 0)))
     ((and (u:number? v) (u:number? u))
      (list (list (u:var v) (* (sym:invert p (poly->number v)) (poly->number u)))
	    (list (u:var v) 0)))
     ((u:number? v) (list (ff:p/c p u (poly->number v)) (list (u:var v) 0)))
     ((< m n)
      (list (list (u:var u) 0) u))
     (else
      (let ((multiplier (sym:invert p lv))
	    (uv (list->vector (cdr u)))
	    (vv (list->vector (cdr v)))
	    (q	(make-vector (+ (- m n) 1) 0))
	    (var (u:var u)))
	(let loop ((k (- m n)))
	  (if (negative? k)
	      (list (ff:unorm p (cons var (vector->list q)))
		    (ff:unorm p (cons var
				      (reverse (list-tail
						(reverse (vector->list uv))
						(+ 1 (- m n)))))))
	      (let ((qk (* (vector-ref uv (+ n k)) multiplier)))
		(vector-set! q k qk)
		(let loop1 ((j (- (+ n k) 1)))
		  (cond ((>= j k)
			 (vector-set! uv j (- (vector-ref uv j)
					      (* qk
						 (vector-ref vv (- j k)))))
			 (loop1 (- j 1)))))
		(loop (- k 1))))))))))

(define (ff:p/p p a b)
  (let ((u (ff:unorm p a))
	(v (ff:unorm p b)))
    (let ((m (ff:degree p u))
	  (n (ff:degree p v)))
      (cond
       ((and (u:zero? u) (u:zero? v))
	(slib:error 'ff:p/p "0/0 is undefined."))
       ((u:zero? v)
	(slib:error 'ff:p/p "Division by 0 is undefined."))
       (else
	(car (ff:p/p->qr p u v)))))))

;;; Given a primitive polynomial a(v) which is an element of a
;;; Unique Factorisation Domain in v, calculate the square free
;;; factorisation of a(v). A primitive polynomial has the content
;;; removed from its coefficients.  The content is the gcd of the
;;; coefficients.
;;; from page 340 of Geddes et al, Algorithms for Computer Algebra
;;;    - mjt
(define (u:sff a)
  (let* ((v (u:var a))
	 (b (poly:diff a v))
	 (c (signcan (poly:gcd a b)))
	 (w (poly:/   a c))
	 (y 0)
	 (z 0)
	 (output '()))
    (poly:sort-factors
     (let loop ((i 1))
       (cond ((not (equal? c 1))
	      (set! y (signcan (poly:gcd w c)))
	      (set! z (poly:/ w y))
;;;	      (ff:print "y = " y " w = " w " c = " c " z = " z)
	      (if (not (number? z))
		  (set! output (if (null? output)
				   (list (list z i))
				   (append output (list (list z i))))))
	      (set! w y)
	      (set! c (poly:/ c y))
;;;	      (ff:print "c = " c)
	      (loop (+ i 1)))
	     (else (append output (list (list w i)))))))))

(define (poly:sff a)
  (if (number? a) (list (list a 1))
      (let* ((v (car a))
	     (b (poly:diff a v))
	     (c (signcan (poly:gcd a b)))
	     (w (poly:/   a c))
	     (y 0)
	     (z 0)
	     (output '()))
	(poly:sort-factors
	 (let loop ((i 1))
	   (cond ((eqv? c 1)
		  (append output (list (list w i))))
		 (else
		  (set! y (signcan (poly:gcd w c)))
		  (set! z (poly:/ w y))
		  (if (not (number? z))
		      (set! output (if (null? output)
				       (list (list z i))
				       (append output (list (list z i))))))
		  (set! w y)
		  (set! c (poly:/ c y))
		  (loop (+ i 1)))))))))

;;;(require (in-vicinity jacal-vicinity "debug"))
;;;(trace poly:sff)

(define (ff:gfroot n p var)
  (let ((d (ff:degree n p)))
    (if (zero? (sym:sym n d))
	(let* ((u (quotient d n))
	       (v (make-vector (+ u 1) 0)))
	  (vector-set! v u 1)
	  (vector-set! v 0 1)
	  (cons var (vector->list v)))
	(slib:error 'ff:gfroot "polynomial has no root." p))))

(define (ff:diff n p v)
  (if (equal? (u:var p) v)
      (do ((i (- (length p) 1) (- i 1))
	   (r '() (cons (sym:sym n (* (- i 1) (list-ref p i))) r)))
	  ((< i 2) (cons v r)))
      (list v 0)))

(define (ff:sff p a)
  (let* ((v (u:var a))
	 (b (ff:diff p a v))
	 (output '()))
    (cond ((not (equal? b (list v 0)))
	   (let* ((c (ff:euclid-gcd p a b))
		  (w (ff:p/p p a c))
		  (y 0)
		  (z 0))
	     (let loop ((i 1))
	       (cond ((not (equal? w (list v 1)))
		      (set! y (ff:euclid-gcd p w c))
		      (set! z (ff:p/p p w y))
		      (if (not (u:one? z))
			  (set! output (if (null? output)
					   (list (list z i))
					   (append output
						   (list (list z i))))))
		      (set! w y)
		      (set! c (ff:p/p p c y))
		      (loop (+ i 1)))))
	     (cond ((not (equal? c (list v 1)))
		    (set! c (ff:gfroot p c v))
		    (set! output (append output
					 (list (list (ff:sff p c) p))))))))
	  (else (set! a (ff:gfroot p a v))
		(set! output (append output (list (list (ff:sff p a) p))))))
    output))

(define (ff:q-matrix p a1)
  (let* ((a (ff:unorm p a1))
	 (n (ff:degree p a))
	 (n-1 (- n 1))
	 (q (make-array 0 n n))
	 (r (make-vector n 0))
	 (r1 (make-vector n 0))
	 (u (* n-1 p))
	 (b (list->vector a)))
    (array-set! q 1 0 0)
    (vector-set! r 0 1)
    (vector-set! r1 0 1)
    (let loop ((i 1))
      (cond ((> i u)
	     q)
	    (else
	     (vector-set! r 0 (* (vector-ref b 1) (- 0 (vector-ref r1 n-1))))
	     (let loop1 ((j 1))
	       (cond ((<= j n-1)
		      (vector-set!
		       r j (sym:sym p (- (vector-ref r1 (- j 1))
				       (sym:sym p (* (vector-ref b (+ j 1))
						   (vector-ref r1 n-1))))))
		      (loop1 (+ j 1)))))
	     (array-map! r1 (lambda (x) x) r)
	     (if (zero? (modular:normalize p i))
		 (array-map! (make-shared-array
			      q (lambda (j) (list (quotient i p) j)) n)
			     (lambda (x) x) r))
;;;		(ff:print i "  r = " r "  q = " q)
	     (loop (+ i 1)))))))

;;; Knuth's null-space-basis, slower but works
(define (ff:null-space-basis p mo var)
  (let* ((n (car (array-dimensions mo)))
	 (m (make-array #f n n))
	 (c (make-vector n -1))
	 (ivec (index-vector n))
	 (result '()))
    (array-copy! mo m)
    (do ((k 0 (+ k 1)))
	((>= k n))
      (let ((j (do ((b 0 (+ b 1)))
		   ((or (>= b n)
			(and (negative? (vector-ref c b))
			     (not (zero? (array-ref m k b))))) b))))
	(if (< j n)
	    (let ((muinv  (* -1 (sym:invert p (array-ref m k j))))
		  (mcolj (make-shared-array m (lambda (a) (list a j)) n)))
	      (array-map! mcolj (lambda (x) (sym:sym p (* x muinv))) mcolj)
	      (vector-set! c j k)
	      (do ((i 0 (+ i 1)))
		  ((>= i n))
		(if (not (= i j))
		    (let ((mcoli (make-shared-array
				  m (lambda (a) (list a i)) n))
			  (mki (array-ref m k i)))
		      (array-map! mcoli
				  (lambda (x y z)
				    (if (>= z k)
				      (sym:sym p (+ x (* mki y)))
				      x))
				  mcoli mcolj ivec)))))
	    (let ((vr (make-vector n 0))
		  (cl (vector->list c)))
	      (array-map! vr
			  (lambda (i)
			    (let ((cl1 (memv i cl)))
			      (cond
				((= i k) 1)
				(cl1 (array-ref m k (- n (length cl1))))
				(else 0))))
			  ivec)
	      (set! result (append result
				   (list (u:lose-lzeros
					   (cons var (vector->list vr))))))))))
    result))

;;; -- > (u:factorz p7)
;     ERROR: vector-ref: Argument out of range 2
;     ;Evaluation took 1934 mSec (601 in gc) 19181 cells work, 6439 bytes other
;     > p7
;     ;Evaluation took 1 mSec (0 in gc) 2 cells work, 31 bytes other
;     (x 1 -3 -1 -3 1 -3 1)
;     > (math)
;     type qed; to return to scheme, type help; for help.
;     e5 : factoruz(1-3*x-x^2-3*x^3+x^4-3*x^5 +x^6);
;     poly = (#(x x #f () #f ()) 1 -3 -1 -3 1 -3 1)
;     e1 = (#(x x #f () #f ()) 1 -3 -1 -3 1 -3 1)
;     Arithmetic Error; Last expression lost.
;;; This problem is caused by berlekamp expecting to find factors when
;;; there are none (ie the polynomial is irreducible).	This problem
;;; comes from the ff:null-space-basis-gcl-bug procedure below.	 This
;;; is the only occurrence of this problem I am aware of.
;;; Fixed by using Knuth's null space basis algorithm above.
;;; Unfortunately, this one is faster.
(define (ff:null-space-basis-gcl-bug p mo var)
  (let* ((n (car (array-dimensions mo)))
	 (m (make-array #f n n)))
    (ff:print " mo = " mo)
    (array-copy! mo m)
    (do ((k 0 (+ k 1)))
	((>= k n))
      (let ((i (do ((b k (+ b 1)))
		   ((or (>= b n) (not (zero? (array-ref m k b)))) b))))
	(if (< i n)
	    (let* ((u	  (array-ref m k i))
		   (uinv  (sym:invert p u))
		   (mcoli (make-shared-array m (lambda (a) (list a i)) n))
		   (mcolk (make-shared-array m (lambda (a) (list a k)) n))
		   (temp  (make-shared-array (make-array #f n n)
					     (lambda (a) (list a 0)) n)))
	      (array-map! mcoli (lambda (x) (sym:sym p (* x uinv))) mcoli)
	      (array-map! temp	(lambda (x) x) mcoli)
	      (array-map! mcoli (lambda (x) x) mcolk)
	      (array-map! mcolk (lambda (x) x) temp)
	      (do ((j 0 (+ j 1)))
		  ((>= j n))
		(if (not (= j k))
		    (let ((mcolj (make-shared-array
				  m (lambda (a) (list a j)) n))
			  (mkj (array-ref m k j)))
		      (array-map! mcolj
				  (lambda (x y)
				    (sym:sym p (- x (sym:sym p (* mkj y)))))
				  mcolj mcolk))))
;;;(ff:print " m = " m)
	      ))))
    (let ((mdiag (make-shared-array m (lambda (a) (list a a)) n)))
      (array-map! mdiag (lambda (x) (sym:sym p (- x 1))) mdiag))
;;;    (ff:print "  After subtraction of 1 along the diagonal,	m = " m)
;;;	 (array-map! m (lambda (x) (sym:sym p (* x -1))) m)
;;;    (ff:print "  After multiplication of -1 along the diagonal,  m = " m)
    (let (;;(i 0)
	  (ret '()))
      (let loop1 ((j 0))
	(if (< j n)
	    (let ((zerow (make-vector n 0)))
	      (let loop2 ()
		(if (< j n)
		    (let ((mrow (make-shared-array
				 m (lambda (a) (list j a)) n)))
		      (cond ((array-equal? mrow zerow)
			     (set! j (+ j 1)) (loop2))))))
	      (if (< j n)
		  (let ((v (make-vector n 0))
			(mrowj (make-shared-array
				m (lambda (a) (list j a)) n)))
		    ;;(set! i (+ i 1))
;;;			 (ff:print " mrowj = " mrowj)
		    (array-copy! mrowj v)
;;;			 (ff:print " v = " v)
		    (set! ret (cons v ret))
		    (loop1 (+ j 1)))))))
;;;	  (ff:print " ret = " ret)
      (map (lambda (x) (ff:unorm p (cons var (vector->list x))))
	   (reverse ret)))))

;;; return an ordered list of elements of Z mod n (symmetric)
(define (ff:generate-field n)
  (let ((v (make-vector n 1))
	(b (sym:bound n)))
    (array-index-map! v (lambda (i) (- i b)))
    (vector->list v)))

;;; return a sorted list of factors of a mod p, where a is square free
;;; and p is a prime number.
(define (ff:berlekamp p a)
  (let* ((q	(ff:q-matrix p a))
	 (n	(car (array-dimensions q)))
	 (var	(u:var a))
	 (qdiag (make-shared-array q (lambda (a) (list a a)) n)))
    (array-map! qdiag (lambda (x) (sym:sym p (- x 1))) qdiag)
    (let* ((vs	(list->vector (ff:null-space-basis p q (u:var a))))
	   (factors (list a))
	   (ffp (ff:generate-field p))
	   (k (vector-length vs)))
      (let loop1 ((r 1))
	(cond ((< (length factors) k)
	       (let loop2 ((us factors))
		 (if (and (< (length factors) k) (not (null? us)))
		     (let ((u (car us))
			   (start #f))
		       (let loop3 ((ss ffp))
			 (if (not (null? ss))
			     (let ((s (car ss)))
			       (if (< (length factors) k)
				   (let ((g (ff:euclid-gcd
					     p (ff:p-p p (vector-ref vs r)
						       (list var s))
					     u)))
				     (cond ((and (not (equal? g u))
						 (not (u:one? g)))
					    (set! factors (delete u factors))
					    (set! u (ff:p/p p u g))
					    (set! factors (delete
							   '()
							   (append factors
								   (list u)
								   (list g))))
					    (set! start #t))
					   (else (loop3 (cdr ss)))))))))
		       (loop2 (if start factors (cdr us))))))
	       (loop1 (+ r 1)))
	      (else (poly:sort-factors factors)))))))

;;; Choose a prime number greater than and relatively prime to n1
(define (choose-prime n1)
  (choose-n (car *primes*) n1 *primes*))

;;; This is silly. I'll change it one day.
(define (choose-n n n1 l)
  (if (and (> n n1) (one? (gcd n n1)))
      n
      (if (null? l)
	  (error "choose-n: Can't find a number larger than " n1)
	  (choose-n (car l) n1 (cdr l)))))

;;; hen:eea-lift computes s,t such that s a + t b = 1 (mod p^k),
;;; deg(s) < deg(b), deg(t) < deg(a), assuming GCD(a mod p, b mod p) =
;;; 1 in Z mod p [x], and that a and b have the same variable and are
;;; univariate.
;;; GCL Algorithm 6.3
;;; This is the extended euclidean gcd algorithm (mod p), but altered
;;; to lift its results to (mod p^k). Arguments a and b are assumed to
;;; be relatively prime.

(define hen:eea-lift
  (memon
   (lambda (a b p k)
     (if (not (equal? (u:var a) (u:var b)))
	 (error "In hen:eea-lift: polys a and b are not in the same variable."))
     (let* ((x (u:var a))
	    (amodp (ff:unorm p a))
	    (bmodp (ff:unorm p b))
	    (g (ff:eea-gcd p amodp bmodp))
	    (s (cadr g))
	    (smodp s)
	    (t (caddr g))
	    (tmodp t)
	    (modulus p))
       (do ((i 1 (+ i 1)))
	   ((>= i k) (let ((res (list s t))) (check-el a b p k res) res))
	 (let* ((e  (number->poly
		     (poly:- (list x 1) (poly:+ (poly:* s a) (poly:* t b)))
		     x))
		(c  (ff:unorm p (u:p/c e modulus)))
		(sb (number->poly (poly:* smodp c) x))
		(tb (number->poly (poly:* tmodp c) x))
		(qr (ff:p/p->qr p sb bmodp))
		(q  (car qr))
		(si (cadr qr))
		(ta (ff:p+p p tb (ff:p*p p q amodp))))
	   (set! s (number->poly (poly:+ s (poly:* si modulus)) x))
	   (set! t (number->poly (poly:+ t (poly:* ta modulus)) x))
	   (set! modulus (* modulus p))))))))

(define (check-el a1 a2 p k result)
  (let* ((x (u:var a1))
	 (p^k (expt p k))
	 (inner-prod (ff:unorm
		      p^k
		      (number->poly
		       (poly:+ (ff:p*p p^k a1 (list-ref result 0))
			       (ff:p*p p^k a2 (list-ref result 1))) x))))
    (cond ((not (equal? (list x 1) inner-prod))
	   (ff:print "In hen:eea-lift: the result is not correct.")
	   (ff:print " a1 = " a1 " a2 = " a2 " result = " result)
	   (ff:print " inner-prod = " inner-prod
		     " supposed result = " (list x 1) " p^k = " p^k)))))


;;; Compute s1..sr such that s(1)*b(1)+...+s(r)*b(r) = 1 (mod p^k)
;;; with deg(s(j)) < deg(a(j))
;;; where b(i) = a(1)*...a(i-1)*a(i+1)*...*a(r),  i = 1..r
;;; Assumes p doesn't divide leading coefft of a(i), i = 1..r
;;;  and the a(i) must each be pairwise relatively prime in Z mod p [x].
;;; GCL alg 6.3
(define hen:multiterm-eea-lift
  (memon (lambda (a p k)
	   (let* ((r (length a))
		  (var (u:var (car a)))
		  (r-1 (- r 1))
		  (r-2 (- r 2))
		  (s	(make-vector r (list var 0)))
		  (q	(make-vector r-1 (list var 0)))
		  (beta (make-vector r (list var 1))))
	     (vector-set! q r-2 (list-ref a r-1))
	     (let loop ((j (- r-2 1)))
	       (cond ((>= j 0)
		      (vector-set! q j (poly:* (list-ref a (+ j 1))
					       (vector-ref q (+ j 1))))
		      (loop (- j 1)))))
	     (let loop ((j 0))
	       (if (< j r-1)
		   (let ((sigma (hen:diophantine
				 (list (vector-ref q j) (list-ref a j))
				 (vector-ref beta j) p k)))
		     (vector-set! beta (+ j 1) (car sigma))
		     (vector-set! s j	 (cadr sigma))
		     (loop (+ j 1)))))
	     (vector-set! s r-1 (vector-ref beta r-1))
	     (check-mel a p k (vector->list s))
;;;(ff:print (vector->list s))
	     (vector->list s)))))

(define (check-mel as p k result)
  (let* ((la (length as))
	 (x (u:var (car as)))
	 (p^k (expt p k))
	 (bs (make-vector la 0)))
;;;    (ff:print bs la)
    (let loop ((i 0))
      (cond ((< i la)
;;;		(ff:print "  i = " i)
	     (let loop1 ((j 0)
			 (acc (list x 1)))
	       (cond ((< j la)
;;;			 (ff:print j bs)
		      (if (not (= i j))
			  (loop1 (+ j 1) (ff:p*p p^k acc (list-ref as j)))
			  (loop1 (+ j 1) acc)))
		     (else (vector-set! bs i acc))))
	     (loop (+ i 1)))))
    (set! bs (vector->list bs))
;;;    (ff:print "  bs = " bs)
    (let ((inner-prod (ff:unorm
		       p^k (number->poly
			    (reduce poly:+ (map (lambda (x y) (ff:p*p p^k x y))
						bs result)) x)))
	  (cn (list x 1)))
;;;	  (ff:print " inner-prod = " inner-prod " x^m = " x^m)
      (cond ((not (equal? cn inner-prod))
	     (ff:print "In hen:multiterm-eea-lift: the result is not correct.")
	     (ff:print " as = " as " bs = " bs " result = " result)
	     (ff:print " inner-prod = " inner-prod " supposed result = " cn
		       " p^k = " p^k))))))


;;; GCL alg 6.3
(define hen:diophantine1
  (memon (lambda (a x m p k)
	   (let ((r (length a))
		 (pk (expt p k))
		 (xm (poly:^ (list x 0 1) m)))
;;;(ff:print "a = " a " x = " x " m = " m " pk = " pk " xm = " xm)
;;;(trace hen:eea-lift hen:multiterm-eea-lift)
	     (if (> r 2)
		 (let ((s (hen:multiterm-eea-lift a p k))
		       (result '()))
		   (let loop ((j 0))
		     (cond ((< j r)
;;;(untrace hen:multiterm-eea-lift)
;;;(ff:print "s" j " = " (list-ref s j) " prod = " (poly:* xm (list-ref s j))
;;;" div = " (ff:p/p->qr (poly:* xm (list-ref s j))
;;;(list-ref a j) pk) " res = " result)
			    (set! result
				  (append
				   result
				   (cdr (ff:p/p->qr
					 pk
					 (poly:* xm (list-ref s j))
					 (list-ref a j)))))
			    (loop (+ j 1)))
			   (else
			    (check-hd1 a x m p k result)
			    result))))
		 (let* ((s (hen:eea-lift (list-ref a 1) (list-ref a 0) p k))
			(q1 (ff:p/p->qr pk (poly:* xm (list-ref s 0))
					(list-ref a 0)))
			(q (car q1))
			(result (list (cadr q1)
				      (ff:p+p pk (poly:* xm (list-ref s 1))
					      (poly:* q (list-ref a 1))))))
		   (check-hd1 a x m p k result)
;;;(ff:print "result = " result)
		   result))))))

(define (check-hd1 as x m p k result)
  (let* ((la (length as))
	 (p^k (expt p k))
	 (bs (make-vector la 0)))
    (let loop ((i 0))
      (cond ((< i la)
	     (let loop1 ((j 0)
			 (acc (list x 1)))
	       (cond ((< j la)
		      (if (not (= i j))
			  (loop1 (+ j 1) (ff:p*p p^k acc (list-ref as j)))
			  (loop1 (+ j 1) acc)))
		     (else (vector-set! bs i acc))))
	     (loop (+ i 1)))))
    (set! bs (vector->list bs))
    (let ((inner-prod
	   (ff:unorm
	    p^k (number->poly
		 (reduce poly:+ (map (lambda (x y) (ff:p*p p^k x y)) bs result))
		 x)))
	  (x^m (number->poly (poly:^ (list x 0 1) m) x)))
      (cond ((not (equal? x^m inner-prod))
	     (ff:print "In hen:diophantine1: the result is not correct.")
	     (ff:print " as = " as " bs = " bs " result = " result)
	     (ff:print " inner-prod = " inner-prod " x^m = " x^m " x = " x
		       " m = " m " p^k = " p^k))))))

;;; GCL alg 6.4
;;; Assumes that the leading coefficient of a given multivariate
;;; polynomial is the last one
(define (hen:mhenseln a i p l u lcu)
;;;(trace hen:multivariate-diophantine)
  (let* ((nu-1 (length i))
	 (nu   (+ nu-1 1))
	 (pl   (expt p l))
	 (av   (make-vector nu a))    ; Partly substituted targets
	 (iv   (make-vector nu '())))  ; Partial ideals for leading coefficients
;;;(ff:print "a = " a " i = " i " p = " p " l = " l " u = " u " lcu = " lcu)
    (let loop ((j (- nu 2)))
      (if (>= j 0)
	(let ((xj     (car (list-ref i j)))
	      (alphaj (cadr (list-ref i j))))
	  (vector-set! iv j (list-tail i j))
	  (vector-set! av j
	      (poly:evaln (vector-ref av (+ j 1)) xj alphaj))
;;;(ff:print "subst j = " j " xj = " xj " alphaj = " alphaj " av = " av)
	  (loop (- j 1)))))
    (let* ((maxdeg (apply max (map (lambda (x) (poly:degree a (car x))) i)))
	   (bu u)
	   (n  (length u)))
      (let loop ((j 0))
;;;(ff:print "j = " j " nu-1 = " nu-1 " maxdeg = " maxdeg " bu = " bu)
;;;(ff:print " av = " av "  iv = " iv)
	(if (< j nu-1)
	  (let* ((bu1 bu)
		 (monomial 1)
		 (e 1)
		 (xj (car (list-ref i j)))
		 (alphaj (cadr (list-ref i j)))
		 (ij (list xj (- alphaj) 1)))
	    (set! bu (lcsubst bu lcu (vector-ref iv (+ j 1)) pl))
;;;(ff:print "bu = " bu " bu1 = " bu1)
	    (set! e (poly:- (vector-ref av (+ j 1)) (reduce poly:* bu)))
;;;(ff:print "e = " e)
	    (let loop1 ((k 1) (monomial ij))
;;;(ff:print "k = " k " av(j+1) = " (poly:degree (vector-ref av (+ j 1)) xj))
	      (if (and (not (eqv? 0 e)) (<= k (poly:degree (vector-ref av (+ j 1)) xj)))
		(let ((c (taylor-coefficient e xj alphaj k)))
;;;(ff:print "monomial = " monomial " c = " c)
;;;(ff:print "bu1 = " bu1 " i = " i)
;;;(ff:print (- (- nu-1 1) j) " small i " (butlast i (- nu-1 j)))
;;;(trace hen:multivariate-diophantine)
		  (if (not (equal? c 0))
		    (let* ((du1 (hen:multivariate-diophantine
				  bu1 (number->poly c (u:var (car bu1)))
				  (butlast i (- nu-1 j)) maxdeg p l))
			   (du (map (lambda (x) (poly:0-degree-norm (poly:* x  monomial))) du1)))
;;;(untrace hen:multivariate-diophantine)
;;;(ff:print "du1 = " du1)
;;;(ff:print "du = " du)
;;;(ff:print "bu = " bu)
		      (set! bu (map (lambda (x y)
				      (ff:mnorm pl (poly:0-degree-norm
						    (poly:+ x y))))
				    bu du))
;;;(ff:print "bu = " bu)
		      (set! e (ff:mnorm pl (poly:0-degree-norm
					    (poly:- (vector-ref av (+ j 1))
						    (reduce poly:* bu)))))))
;;;(ff:print "e = " e)
		(loop1 (+ k 1) (poly:* monomial ij)))))
	    (loop (+ j 1)))))
;;;(ff:print "bu product = " (reduce-init poly:* 1 bu))
;;;(ff:print "bu = " bu)
;;;(ff:print "a = " a)
      (if (equal? a (poly:0-degree-norm (reduce-init poly:* 1 bu))) bu #f))))

;;; Substitute an updated leading coefficient into each member of u.
;;; The member of the correct leading coefficients in lcu corresponding to
;;; a factor in u, is reduced mod the ideal id.
(define (lcsubst u lcu id pl)
;;;(ff:print "u = " u " lcu = " lcu " id = " id)
  (let* ((save	 (map (lambda (x) (butlast x 1)) u))
	 (unsubs (map (lambda (x y) (append x (list y))) save lcu)))
;;;(ff:print "save = " save "  unsubs = " unsubs)
    (map (lambda (x)
	   (let loop ((i id))
;;;(ff:print " i = " i " x = " x)
	     (cond ((null? i) x)
		   (else (set! x (ff:mnorm
				  pl (poly:evaln x (caar i) (cadar i))))
			 (loop (cdr i))))))
	 unsubs)))

;;; GCL alg 6.2 partly tested
(define (hen:multivariate-diophantine a c i d p k)
;;;(trace hen:diophantine1 hen:eea-lift)
  (let* ((r	  (length a))
	 (nu-1	  (length i))
	 (nu	  (+ nu-1 1))
	 (pk	  (expt p k)))
;;;(ff:print "r = " r " nu = " nu " pk = " pk " d = " d)
;;;(ff:print " factors = " a " target = " c " i = " i)
    (if (> nu 1)
	(let* ((ap (reduce-init poly:* 1 a))
	       (b	(map (lambda (x) (poly:/ ap x)) a))
	       (xnu	(car (list-ref i (- nu-1 1))))
	       (alphanu (cadr (list-ref i (- nu-1 1))))
	       (anew (map (lambda (x) (poly:evaln x xnu alphanu)) a))
	       (cnew (poly:evaln c xnu alphanu))
	       (inew (del (list xnu alphanu) i '()))
	       (sigma (hen:multivariate-diophantine anew cnew inew d p k))
	       (e (ff:mnorm
		   pk (poly:- c (reduce-init poly:+ 0 (map poly:* sigma b)))))
	       (monomial 1))
;;;(ff:print "b = " b " anew = " anew " cnew = " cnew " inew = " inew " sigma = " sigma " e = " e)
	  (let loop ((m 1))
;;;(ff:print "m = " m " d = " d " e = " e)
	    (cond
	     ((and (<= m d) (not (poly:0? e)))
;;;(ff:print "monomial = " monomial " xnu = " xnu " alphanu = " alphanu)
	      (set! monomial (poly:* monomial (poly:- (list xnu 0 1) alphanu)))
	      (let ((cm (taylor-coefficient e xnu alphanu m)))
;;;(ff:print "Taylor cm " cm)
		(if (not (equal? cm 0))
		    (let ((ds (hen:multivariate-diophantine anew cm inew d p k)))
		      (set! ds (map (lambda (x1) (poly:* x1 monomial)) ds))
		      (set! sigma (map poly:+ sigma ds))
;;;(ff:print "ds = " ds " sigma = " sigma "e = " e)
		      (set! e (ff:mnorm
			       pk (poly:- e
				       (reduce-init poly:+ 0
						    (map poly:* ds b)))))
;;;(ff:print "ds = " ds " sigma = " sigma "e = " e " pk = " pk)
		      )))
	      (loop (+ m 1)))
	     (else
	      (map (lambda (x) (ff:mnorm pk x)) sigma)))))
	(let* ((x1      (u:var (car a)))
	       (sigma   (make-list r 0))
	       (c1      (number->poly c x1)) ;;; So we can deconstruct consistently
	       (dc      (poly:degree c1 x1)))
;;;(ff:print " a = " a " c1 = " c1 " sigma = " sigma)
	  (let loop ((m 1))
;;;(ff:print " m = " m " dc = " dc)
	    (if (<= (- m 1) dc)
		(let* ((cm (leading-coeff (list-ref c1 m) x1))
;;; IF cm = 0 then we don't need to do the diophantine solution.
		       (ds (hen:diophantine1 a x1 (- m 1) p k)))
		  (check-hd1 a x1 (- m 1) p k ds)
;;;(ff:print " cm = " cm " ds = " ds)
		  (set! ds (map (lambda (x) (u:p*c x cm)) ds))
		  (set! sigma (map (lambda (x y)
				     (number->poly (poly:+ x y) x1))
				   sigma ds))
;;;(ff:print " sigma =" sigma " ds =" ds)
		  (loop (+ m 1)))))
;;;(untrace hen:diophantine1 hen:eea-lift)
	  (map (lambda (x) (ff:unorm pk x)) sigma)))))

;;; Don't know if this one is OK, taken from poly.scm, commented out there.
(define (leading-coeff p var)
  (poly:coeff p var (poly:degree p var)))

;;; mth Taylor series coefficient in expansion about var = n
(define (taylor-coefficient poly var n m)
  (poly:/
    (poly:evaln (let loop ((i 1) (fn poly))
		  (if (> i m) fn (loop (+ 1 i) (poly:diff fn var))))
		var n)
    (factorial m)))

;;; Evaluate simple univariate polynomials the naive way
(define (u:evaln p n)
 (let ((d+1 (+ 1 (u:degree p))))
  (let loop ((i 2) (i-1 1) (acc (list-ref p 1)))
    (if (> i d+1)
      acc
      (loop (+ i 1) i (+ acc (* (list-ref p i) (expt n i-1))))))))

;;; Substitute n into a univariate polynomial p, without assuming
;;; that p is simple numbers in the coefficients.
(define (u:evalp p n)
 (let ((d+1 (+ 1 (u:degree p))))
  (let loop ((i 2) (i-1 1) (acc (list-ref p 1)))
    (if (> i d+1)
      acc
      (loop (+ i 1) i (poly:+ acc (poly:* (list-ref p i) (poly:^ n i-1))))))))

;;; Evaluate a polynomial, p, substituting a number n for the variable var
;;; Do some normalisation along the way
(define (poly:evaln p var n)
;;;(ff:print "p " p " var " var " n " n)
  (poly:0-degree-norm
    (cond
      ((number? p) p)
      ((u:number? p) (cadr p))
      ((equal? (car p) var)
	(if (poly:univariate? p)
	  (u:evaln p n)
	  (reduce poly:+
		  (map (lambda (x y) (if (number? x)
				       (* x (expt n y))
				       (poly:* (poly:evaln x var n) (expt n y))))
		       (cdr p)
		       (power-list p var)))))
      (else
	(cons (car p)
	      (map (lambda (x) (if (number? x)
				 x (poly:evaln x var n)))
		   (cdr p)))))))

;;; evaluate a polynomial p mod the ideal i
;;; (An evaluation homomorphism)
(define (poly:eval-ideal p i)
  (if (null? i) p (poly:eval-ideal (poly:evaln p (caar i) (cadar i)) (cdr i))))

;;; generate the powers of x for a poly
(define (power-list p v)
  (let ((vec (make-vector (+ 1 (poly:degree p v)) 1)))
    (array-index-map! vec (lambda (i) i))
    (vector->list vec)))

(define (index-list n)
  (let ((vec (make-vector n 1)))
    (array-index-map! vec (lambda (i) i))
    (vector->list vec)))

(define (index-vector n)
  (let ((vec (make-vector n 1)))
    (array-index-map! vec (lambda (i) i))
    vec))

;;; memoised diophantine equation solver is much faster
(define hen:diophantine
  (memon (lambda (a c p k)
	   (let* ((r	   (length a))
		  (p^k	   (expt p k))
		  (x1	   (u:var c))
		  (dc	   (u:degree c))
		  (sigma   (make-list r 0)))
	     (let loop ((m 1))
	       (if (<= (- m 1) dc)
		   (let* ((cm (list-ref c m))
			  (ds (hen:diophantine1 a x1 (- m 1) p k)))
		     (set! ds (map (lambda (x) (u:p*c x cm)) ds))
		     (set! sigma (map (lambda (x y)
					(number->poly (poly:+ x y) x1))
				      sigma ds))
		     (loop (+ m 1)))
		   (let ((result (map (lambda (x) (ff:unorm p^k x)) sigma)))
		     (check-hd a c p k result)
		     result)))))))

(define (check-hd as c p k result)
  (let* ((la (length as))
	 (x (u:var c))
	 (p^k (expt p k))
	 (bs (make-vector la 0)))
    (let loop ((i 0))
      (cond ((< i la)
	     (let loop1 ((j 0)
			 (acc (list x 1)))
	       (cond ((< j la)
		      (if (not (= i j))
			  (loop1 (+ j 1) (ff:p*p p^k acc (list-ref as j)))
			  (loop1 (+ j 1) acc)))
		     (else (vector-set! bs i acc))))
	     (loop (+ i 1)))))
    (set! bs (vector->list bs))
    (let ((inner-prod (ff:unorm
		       p^k (number->poly
			    (reduce poly:+ (map (lambda (x y) (ff:p*p p^k x y))
						bs result)) x)))
	  (cn (ff:unorm p^k c)))
      (cond ((not (equal? cn inner-prod))
	     (ff:print "In hen:diophantine: the result is not correct.")
	     (ff:print " as = " as " bs = " bs " result = " result)
	     (ff:print " inner-prod = " inner-prod " cn = " cn " p^k = " p^k)
	     )))))

;;; 2 factor univariate linear Hensel lifting algorithm
(define (hen:ulhensel2 f g1 h1 p k)
  (let* ((g g1)
	 (h h1)
	 (lf (u:leading f))
	 (ghrecip (ff:eea-gcd p g1 h1))
	 (grecip (cadr ghrecip))
	 (hrecip (caddr ghrecip))
	 (x (u:var f)))
    (let loop ((i 2)
	       (modulus p)
	       (modulus*p (* p p)))
      (if (<= i k)
	  (let* ((discrepancy
		  (u:p/c (ff:p-p modulus*p (ff:p/c modulus*p f lf)
				 (number->poly (poly:* g h) x))
			 modulus))
		 (gc (cadr (ff:p/p->qr
			    p (number->poly (poly:* hrecip discrepancy) x) g1)))
		 (hc (cadr (ff:p/p->qr
			    p (number->poly (poly:* grecip discrepancy) x)
			    h1))))
	    (set! g (poly:+ g (poly:* gc modulus)))
	    (set! h (poly:+ h (poly:* hc modulus)))
	    (loop (+ i 1) modulus*p (* modulus*p p)))
	  (list g h)))))

;;; Experimental (slow) 3 factor univariate linear Hensel lifting algorithm
(define (hen:ulhensel3 f g1 h1 i1 p k)
  (let* ((g g1)
	 (h h1)
	 (lf (u:leading f))
	 (i i1)
	 (x (u:var f)))
;;;    (ff:print g h i)
    (let loop ((l 2)
	       (modulus p)
	       (modulus*p (* p p)))
      (if (<= l k)
	  (let* ((diff
		  (u:p/c (ff:p-p modulus*p (ff:p/c modulus*p f lf)
				 (number->poly (poly:* (poly:* g h) i) x))
			 modulus))
		 (dghi (hen:diophantine (list g1 h1 i1) diff p 1))
;;;		     (dghi (hen:diophantine (list g1 h1 i1) diff p (- l 1)))
		 (dg (car dghi))
		 (dh (cadr dghi))
		 (di (caddr dghi)))
	    (set! g (poly:+ g (poly:* dg modulus)))
	    (set! h (poly:+ h (poly:* dh modulus)))
	    (set! i (poly:+ i (poly:* di modulus)))
;;;		(ff:print diff " dg = " dg " dh = " dh " di = " di g h i)
	    (loop (+ l 1) modulus*p (* modulus*p p)))
	  (list g h i)))))

;;; Experimental (slow) n factor univariate linear Hensel lifting algorithm
(define (hen:ulhenseln f as p k)
  (let ((x (u:var f))
	(lf (u:leading f)))
;;;    (ff:print f as)
    (let loop ((l 2)
	       (modulus p)
	       (modulus*p (* p p)))
      (if (<= l k)
	  (let* ((diff
		  (u:p/c (ff:p-p modulus*p (ff:p/c modulus*p f lf)
				 (number->poly (reduce poly:* as) x)) modulus))
		 (ds (hen:diophantine as diff p 1)))
	    (set! as (map (lambda (x y) (poly:+ x (poly:* y modulus))) as ds))
;;;	    (ff:print diff " ds = " ds " as = " as l)
	    (loop (+ l 1) modulus*p (* modulus*p p)))
	  as))))

;;; Return the Landau-Mignotte Bound of the coefficients of f
(define (landau-mignotte-bound f)
  (abs (* (sqrt (apply + (map (lambda (x) (* x x)) (cdr f))))
	  (expt 2 (u:degree f))
	  (u:leading f))))

;;; Return the variable of maximum degree in poly
(define (max-degree-var poly)
  (let* ((vars (poly:vars poly))
	 (nvars (length vars))
	 (degs (map (lambda (x) (poly:degree poly x)) vars))
	 (maxdeg (apply max degs)))
    (list-ref vars (- nvars (length (memv maxdeg degs))))))

;;; Return the variable with content of maximum degree in poly
(define (max-content-degree-var poly)
  (let* ((vars (poly:vars poly))
	 (nvars (length vars))
	 (conts (map (lambda (x) (poly:contz poly x)) vars))
	 (degs	(map (lambda (y)
		       (apply max (map (lambda (x) (poly:degree y x)) vars)))
		     conts))
	 (maxdeg (apply max degs)))
    (list-ref vars (- nvars (length (memv maxdeg degs))))))

;;; Special var power factors of poly
(define (svpf poly)
  (let loop ((p poly) (n 0))
    (if (number? (cadr p))
      (if (zero? (cadr p))
	(if (null? (cddr p))
	  (+ n 1)
	  (loop (cons (car p) (cddr p)) (+ n 1)))
	n) n)))

;;; Factorise multivariate polynomial over Z (sorted)
(define (poly:factorz poly)
;;;(trace poly:factorszpp hen:mhenseln poly:sff)
  (let* ((vars (poly:vars poly))
	 (nvars (length vars)))
    (cond
     ((= nvars 0) (list (list (list poly) 1))) ; number
     ((= nvars 1)			; univariate
      (u:factorz poly))
     (else				; multivariate
      (let* ((spec-var (car poly))
	     (spec-var-poly (list spec-var 0 1))
	     (psign (poly:unitz poly spec-var))
	     (pcont (poly:contz poly spec-var))
	     (ppoly1 (poly:primz poly spec-var))
	     (spec-var-pf (svpf ppoly1))
	     (ppoly (poly:/ ppoly1
			    (poly:^ spec-var-poly spec-var-pf))))
;;;(ff:print "spec-var = " spec-var "  psign = " psign " pcont = " pcont
;;;	     " ppoly = " ppoly " spec-var-pf = " spec-var-pf " ppoly1 = " ppoly1)
	(let*((pcontfactors (poly:factorz pcont))
	      (ppolyfactors (cond
			     ((zero? spec-var-pf)
			      (map (lambda (x)
				     (list (poly:sort-factors
					    (poly:factorszpp (car x)))
					   (cadr x))) (poly:sff ppoly)))
			     ((number? ppoly)
			      (list (list (list spec-var-poly) spec-var-pf)))
			     (else
			      (cons
			       (list (list spec-var-poly) spec-var-pf)
			       (map (lambda (x) (list (poly:sort-factors
						       (poly:factorszpp (car x)))
						      (cadr x))) (poly:sff ppoly)))))))
;;;(ff:print "spec-var = " spec-var "  psign = " psign " pcont = " pcont
;;;	     " ppoly = " ppoly)
;;;(ff:print " pcontfactors = " pcontfactors " ppolyfactors = " ppolyfactors)
	  (cond
	   ((and (equal? pcont 1) (equal? psign 1))
	    ppolyfactors)
	   ((number? pcont)
	    (cons (list (list (* pcont psign)) 1) ppolyfactors))
	   ((equal? psign 1)
	    (poly:sort-factors
	     (compact-powers
	      (append pcontfactors ppolyfactors))))
	   (else
	    (poly:sort-factors
	     (compact-powers
	      (append (list (list (list psign) 1))
		      pcontfactors ppolyfactors)))))))))))

;;;(trace poly:factorz)

(define (compact-numbers l)
;;;(ff:print "cn l = " l)
  (let ((c 1)
	(fs '()))
    (for-each (lambda (x)
		(if (number? x)
		  (set! c (* c x))
		  (set! fs (cons x fs)))) l)
;;;(ff:print (list (list c 1)) "  " (reverse fs))
    (if (= c 1) fs
      (cons c (reverse fs)))))

(define (compact-powers l)
;;;(ff:print "cp l = " l)
  (let ((fs '()))
    (for-each (lambda (x)
		(let ((f (assq (cadr x) fs)))
		  (if f
		    (set! fs (cons (cons (car f)
					 (append (car x) (cdr f)))
				   (del f fs '())))
		    (set! fs (cons (cons (cadr x) (car x)) fs)))))
	      l)
;;;(ff:print "cp fs = " fs)
    (map (lambda (x)
	   (cons (compact-numbers
		   (poly:sort-factors (cdr x))) (list (car x)))) fs)))

(define (unlucky? maxdeg poly1 spec-var)
  (or (not (= maxdeg (poly:degree poly1 spec-var)))
      (not (eqv? 1 (poly:gcd (poly:diff poly1 spec-var) poly1)))))

;;; Factorise square-free primitive multivariate polynomial over Z (sorted)
(define (poly:factorszpp ppoly)
;;;(ff:print "ppoly = " ppoly)
  (if (number? ppoly)
      ppoly
      (let* ((vars (poly:vars ppoly))
	     (spec-var (car ppoly))
	     (maxdeg (poly:degree ppoly spec-var))
	     (ivars (del spec-var vars '()))
	     (nivars (length ivars))
	     (lcpoly (leading-coeff ppoly spec-var))
	     (lcfactors (if (number? lcpoly)
			    (factor lcpoly)
;;;			   (map (lambda (x) (list x 0 1))
;;;				(poly:vars lcpoly))
			    (strip-powers (poly:factorz lcpoly) '())))
	     (lcnums (apply append (map (lambda (x) (if (number? x) (list x) '())) lcfactors)))
	     (nlcnums (length lcnums))
	     (nlcfs (length lcfactors)))
	(let loop ((start 1))
;;;(ff:print "start = " start)
	  (let* ((image ppoly)
		 (iu
		  (let loop ((mr+1 (+ nivars nlcnums 2))
			     (random-ideal (make-random-ideal ivars lcnums (+ nivars nlcnums 3))))
;;;(ff:print ".")
		    (let* ((poly1 (poly:eval-ideal image random-ideal))
			   (elcfactors
			    (map (lambda (x)
				   (remove-duplicates
				    (factor (poly:eval-ideal x random-ideal)) '() ))
				 lcfactors))
			   (us (uniques elcfactors)))
;;;(ff:print image "  " poly1 "	 " maxdeg "  " spec-var "  " elcfactors "  "
;;;  lcfactors "  " us "  " random-ideal)
;;;(ff:print "factors = " (poly:factorz lcpoly))
;;;(ff:print "lcfactors = " lcfactors " elcfactors = " elcfactors "  us = " us " lcnums = " lcnums)
		      (cond
		       ((or (unlucky? maxdeg poly1 spec-var) (some null? us))
			(loop (+ mr+1 1) (make-random-ideal ivars lcnums mr+1)))
		       (else
			(set! image poly1)
			(list random-ideal
			      (if (eqv? 1 lcpoly)
				  (make-list nlcfs '(1 1))
				  (map (lambda (x y) (list x y))
				       (map car us) lcfactors))))))))
		 (ideal (car iu))
		 (uniqs (cadr iu))
		 (p (let loop ((cp 3))
;;;(ff:print image "  " cp)
		      (if (or (not (= maxdeg (ff:degree cp image)))
			      (not (every (lambda (x) (eqv? x 1))
					  (map cadr (ff:sff cp image)))))
			  (loop (car (primes:primes> (+ cp 1) 1))) cp)))
		 (image-factors
		  (let* ((slpoly (sign (u:leading image)))
			 (cpoly	 (u:contz image))
			 (factors (poly:sort-factors (u:factorsz image))))
		    (if (and (= cpoly 1) (= slpoly 1))
			factors
			(cons (list spec-var (* slpoly cpoly)) factors))))
		 (imagelcs (map u:leading image-factors))
		 (k (u:prime-power image p)))
;;;(ff:print "image-factors = " image-factors " ideal = " ideal " uniqs = " uniqs)
;;;(ff:print "imagelcs = " imagelcs " lcfactors	 = " lcfactors)
;;;(ff:print "clcs = " (correct-lcs uniqs imagelcs lcpoly))
	    (if (= (length image-factors) 1)
		(list ppoly)
		(let* ((clcs (correct-lcs uniqs imagelcs lcpoly))
		       (res (hen:mhenseln ppoly ideal p k image-factors clcs)))
		  (if (not res)
		      (loop (+ start 1))
		      res))))))))

;;; Make a random ideal for the list of variables ivars with the
;;; random numbers > 1 and <= r.  Each variable must have a unique
;;; value to assist in sorting out leading coefficients.
;;; r > (length ivars) + (length nums)
;;; nums is a list of numbers which cannot appear in the ideal.
;;; The random number must be greater than one to enable the recognition of
;;; multiple factors in the leading coefficient.
(define (make-random-ideal ivars nums r)
  (if (> (+ (length ivars) (length nums)) r)
    (error "In make-random-ideal: r is too small." ivars nums r))
  (map (lambda (x y) (list x y)) ivars (mri ivars '() nums r)))

(define (mri ivars sofar nums r)
  (if (null? ivars)
    sofar
    (let ((rn (let loop ((rn1 (+ 2 (random r))))
;;;(ff:print ".")
		(if (member rn1 nums)
		  (loop (+ 2 (random r))) rn1))))
      (mri (cdr ivars) (cons rn sofar) (cons rn nums) r))))

;;; Strip a list of factors and powers of the powers
(define (strip-powers l acc)
  (if (null? l) acc (strip-powers (cdr l) (append (caar l) acc))))

;;; Return a suitable power for the prime p to be raised to, to
;;; cover all possible coefficients of poly and its factors.
(define (u:prime-power poly p)
  (filog p (inexact->exact (ceiling (* 2 (landau-mignotte-bound poly))))))

;;; Add '(1 1)s to the list of unique factors (uns) if there are more image
;;; factors (ifs) than uns
(define (add-ones uns ifs)
  (let ((luns (length uns))
	(lifs (length ifs)))
    (if (> lifs luns)
      (append (make-list (- lifs luns) '(1 1)) uns)
      uns)))

;;; Generate the list of correct leading coefficients
;;; uniqs - association list of a unique prime factor for each leading
;;; coefficient of a polynomial's factors, evaluated over some ideal and their
;;; corresponding unevaluated multivariate factors.
;;; iflcs - factors of the leading coefficient of the univariate image of the
;;; polynomial, evaluated over the same ideal.

(define (correct-lcs uniqs imagelcs lc)
;;;(ff:print "uniqs = "uniqs " imagelcs = " imagelcs " lc = " lc)
  (map
   (lambda (current-imagelc)
     (let ((res 1)
	   (lun (length uniqs)))
       (let loop ((i 0))
;;;(ff:print "loop starts i = " i " lc = " lc " uniqs length = " lun " uniqs " uniqs)
	 (cond ((and (not (equal? 1 lc)) (< i lun))
		(let* ((current (list-ref uniqs i))
		       (unique-prime (car current))
		       (fact	   (cadr current)))
;;;(ff:print "current " current "  current-imagelc " current-imagelc "	i " i)
		  (let loop1 ()
		    (cond ((and (zero? (remainder current-imagelc unique-prime))
				(poly:/? lc fact))
			   (set! res (poly:* fact res))
			   (set! current-imagelc (quotient current-imagelc unique-prime))
			   (set! lc (poly:/ lc fact))
;;;(ff:print "res " res "  uniqs " uniqs)
			   (loop1)))
		    (loop (+ i 1)))))
	       (else
;;;(ff:print "res " res)
		res)))))
   imagelcs))

;;;(trace correct-lcs)

;;; Factorise a square-free univariate polynomial over the integers
(define (u:factorsz f1)
  (let* ((df (u:degree f1))
	 (var (u:var f1))
	 (lf  (u:leading f1))
	 (pfsp (let loop ((p 3))
		 (if (or  (not (= df (ff:degree p f1)))
			  (not (u:one? (ff:euclid-gcd p (ff:diff p f1 var)
						      (ff:unorm p f1)))))
		     (loop (car (primes:primes> (+ p 1) 1)))
		     (list (ff:berlekamp p (ff:make-monic p f1)) p))))
	 (pfs (car pfsp))
	 (p (cadr pfsp))
	 (n (length pfs))
	 (f (ff:make-monic p f1))
	 (answer '()))
    (cond
     ((= n 1) (list f1))
     ((= n 2)
      (let* ((k (u:prime-power f1 p))
	     (factors (hen:ulhensel2 f1 (car pfs) (cadr pfs) p k))
	     (pk (expt p k))
	     (fs factors) (g '()))
	(for-each (lambda (x)
		    (cond ((eqv? 0 (cadr (poly:pdiv (poly:* f1 lf)
						    (ff:p*c pk x lf) var)))
			   (set! answer (cons (u:primz (ff:p*c pk x lf)) answer))
			   (set! factors (delete	 x factors))
			   (set! n (- n 1)))))
		  fs)
	(if (not (null? factors))
	    (list f1)
	    answer)))
     ((= n 3)
      (let* ((k (u:prime-power f1 p))
	     (factors (hen:ulhensel3 f1 (car pfs) (cadr pfs) (caddr pfs) p k))
	     (fs factors)
	     (pk (expt p k))
	     (g '()))
	(for-each (lambda (x)
		    (cond ((eqv? 0 (cadr (poly:pdiv (poly:* f1 lf)
						    (ff:p*c pk x lf) var)))
			   (set! answer (cons (u:primz (ff:p*c pk x lf)) answer))
			   (set! factors (delete	 x factors))
			   (set! n (- n 1)))))
		  fs)
	(if (not (null? factors))
	    (cons (u:primz (ff:p*c pk (ff:unorm pk (reduce-init
						    poly:*
						    (list var 1) factors)) lf))
		  answer)
	    answer)))
     (else
      (let* ((k (u:prime-power f1 p))
	     (factors (hen:ulhenseln f1 pfs p k))
	     (fs factors)
	     (pk (expt p k))
	     (e '())
	     (g '()))
	(for-each (lambda (x)
		    (cond ((eqv? 0 (cadr (poly:pdiv
					  (poly:* f1 lf) (ff:p*c pk x lf) var)))
			   (set! answer (cons (u:primz (ff:p*c pk x lf)) answer))
			   (set! factors (delete	 x factors))
			   (set! n (- n 1)))))
		  fs)
	(call-with-current-continuation
	 (lambda (exit)
	   (let loop ((combine 2))
	     (if (<= (* 2 combine) n)
		 (let ((fs factors))
		   (set! e (combinations factors combine))
		   (for-each
		    (lambda (u)
		      (set! g (ff:unorm (expt p k) (reduce poly:* u)))
		      (cond
		       ((eqv? 0 (cadr (poly:pdiv f1 (ff:p*c pk g lf) var)))
			(set! answer (cons (u:primz (ff:p*c pk g lf)) answer))
			(for-each
			 (lambda (x)
			   (set! factors (delete x factors)))
			 u)
			(set! n (- n combine))
			(if (> (* 2 combine) n) (exit #t)))))
		    e)
		   (loop (+ combine 1)))))))
	(if (not (null? factors))
	    (cons (u:primz (ff:p*c pk (reduce-init
				       poly:* (list var 1) factors) lf)) answer)
	    answer))))))

;;; return a sorted list of factors of the univariate polynomial poly
;;; over the integers (Z)
(define (u:factorz poly)
  (if (number? poly)
    (list (list poly 1))
    (let* ((slpoly (u:unitz poly))
	   (cpoly  (u:contz poly))
	   (factors (map (lambda (x) (list (poly:sort-factors
					     (u:factorsz (car x)))
				     (cadr x)))
			 (u:sff (u:primz poly)))))
      (cond
	((and (= cpoly 1) (= slpoly 1))	 factors)
	(else (cons (list (list (* slpoly cpoly)) 1) factors))))))

;;; Least common multiple
(define (poly:lcm x)
   (poly:/ (reduce poly:* x)
	   (reduce poly:gcd x)))

;;; multiply powers of a list of factors l, by n
(define (mult-power l n)
   (map (lambda (x) (list (car x) (* n (cadr x)))) l))

;;; Factorise over the Rationals (Q)
(define (u:factorq poly)
  (cond
    ((rat? poly)
      (append (u:factorz (num poly)) (mult-power (u:factorz (denom poly)) -1)))
    (else
      (u:factorz poly))))

(define (poly:factorq poly)
  (cond
    ((rat? poly)
      (append (poly:factorz (num poly))
	      (mult-power (poly:factorz (denom poly)) -1)))
    (else
      (poly:factorz poly))))

;;; Partial fraction expansion of a rational univariate polynomial
;;; The denominator, dr, must be square free.
(define (u:partial-fraction-expand nr dr)
  (let* ((drfs1 (u:factorq dr))
	 (drfs (remove-exponents drfs1 '()))
	 (p 3)
	 (k 5)
	 (ss   (hen:diophantine drfs (list (u:var (car drfs)) 1) p k)))
    (let ((res (map (lambda (x y) (make-rat (poly:* nr x) y)) ss drfs)))
      (ff:print res)
      res)))

(define (remove-exponents fs fs1)
  (cond
    ((null? fs) fs1)
    ((number? (caar fs))
      (ff:print "n " (caar fs))
      (remove-exponents (cdr fs) (fs1)))
    (else
      (ff:print "e " (caar fs))
      (remove-exponents (cdr fs) (append (caar fs) fs1)))))

(define (ff:check-arg e)
  (cond ((not (poly:univariate? e))
	 (bltn:error 'Not-a-univariate-polynomial e))))

(define (ff:check-poly-arg e)
  (cond ((not (poly:multivariate? e))
	 (bltn:error 'Not-a-multivariate-polynomial e))))

(define (ff:check-prime n)
  (cond ((not (memv n *primes*))
	 (bltn:error 'Not-a-Prime-Number n))))

;;; JACAL's builtin functions.
(defbltn 'sff
  (lambda (poly)
    (let ((e (licit->polxpr poly)))
      (if (not (eqv? 1 (poly:contz e (car e))))
	(bltn:error 'Not-a-primitive-polynomial poly)
	(poly:sff e)))))

(defbltn 'usff
  (lambda (poly)
    (let ((e (licit->polxpr poly)))
      (ff:check-arg e)
      (cond
       ((not (equal? e (u:primz e)))	   ;this test should be replaced
	(bltn:error 'Not-a-primitive-polynomial poly))
       (else (u:sff e))))))

(defbltn 'ffsff
  (lambda (poly pn . k1)
    (let ((p (licit->polxpr poly))
	  (n (licit->polxpr pn))
	  (k (licit->polxpr (if (null? k1) 1 (car k1)))))
      (ff:check-arg p)
      (ff:check-prime n)
      (cond
       ((not (ff:monic? n p))
	(bltn:error 'Not-monic-mod-n p))
       ((not (> k 0))
	(bltn:error 'Not-greater-than-zero k))
       (else (ff:sff (expt n k) p))))))

(defbltn 'berl
  (lambda (poly pn)
    (let ((p (licit->polxpr poly))
	  (n (licit->polxpr pn)))
      (ff:check-arg p)
      (ff:check-prime n)
      (cond
       ((not (ff:monic? n p))
	(bltn:error 'Not-monic-mod-n p))
       ((not (= (u:degree p) (ff:degree n p)))
	(bltn:error 'Not-same-degree-when-reduced-mod-n (ff:norm p n)))
       (else (ff:berlekamp n p))))))

(defbltn 'factoruz
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (cond
	((rat? e1)
	  (bltn:error 'Rational-in-factoruz poly))
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	(else (ff:check-arg e1) (u:factorz  e1))))))

;;; Should have R equivalents for unitz, contz, primz.
;;; These return the unit (1 or -1), content (normalised to positive
;;; leading number coefficient) and primitive (likewise normalised)
;;; parts of a multivariate polynomial.
(defbltn 'unitz
  (lambda (poly var)
    (let ((e1 (expr:normalize poly))
	  (e2 (u:var (expr:normalize var))))
      (cond
	((rat? e1)
	  (bltn:error 'Rational-in-u poly))
	((number? e1) (sign e1))
	(else (poly:unitz e1 e2))))))

(defbltn 'contz
  (lambda (poly var)
    (let ((e1 (expr:normalize poly))
	  (e2 (u:var (expr:normalize var))))
      (cond
	((rat? e1)
	  (bltn:error 'Rational-in-content poly))
	((number? e1) (abs e1))
	(else (poly:contz e1 e2))))))

(defbltn 'primz
  (lambda (poly var)
    (let ((e1 (expr:normalize poly))
	  (e2 (u:var (expr:normalize var))))
      (cond
	((rat? e1)
	  (bltn:error 'Rational-in-content poly))
	((number? e1) 1)
	(else (poly:primz e1 e2))))))

(defbltn 'factoruq
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (cond
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	(else (ff:check-arg e1)
	      (u:factorq e1))))))

(defbltn 'factorq
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (cond
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	((poly:univariate? e1) (u:factorq e1))
	(else (ff:check-poly-arg e1)
	  (let ((result
	      (poly:factorq e1)))
;;;(ff:print "result = " result )
	    result))))))

(defbltn 'factorz
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (cond
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	((not (list?  e1))
	  (bltn:error 'Not-a-polynomial poly))
	((not (poly:var? (car e1)))
	  (bltn:error 'Not-a-polynomial poly))
	(else
	  (let ((res (poly:factorz e1)))
	    (ff:print res) res))))))

(defbltn 'parfrac
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (u:partial-fraction-expand (num e1) (denom e1)))))

;;; (x+1)/(x+2)
;;;(#(:@ :@ #f () #f ()) (#(x x #f () #f ()) -1 -1) (#(x x #f () #f ()) 2 1))

(define a '(x -5 2 8 -3 -3 0 1 0 1))
(define b '(x 21 -9 -4 0 5 0 3))
(define b1 '(x -2 -9 -4 0 5 0 3))
(define c '(x 1 1 2 2 1 1))
(define d '(x 1 0 2 0 1))
(define e '(x 2 1 4 2 2 1))
(define f '(x -36 42 -84 48))
(define g '(x -30 44 -10 -4))
(define h '(x -5 2 1 -3 -3 0 1 0 1))
(define i '(x 21 -9 -4 0 5 0 3))

(define j '(x -5 2 8 -3 -3 0 1 0 1))	; page 280
(define k '(x -2 -9 -4 0 5 0 3))	; page 280 gcd 21 mod 23

(define l '(x -360 -171 145 25 1))
(define m '(x -15 -14 -1 15 14 1))	; page 302 mod 5 = x^4 + 4x
					;	   mod 7 = x^2 + 1
					;	   mod 11 = x^2 + 3x + 4
(define o '(x 1 0 2 2 0 1 1 0 2 2 0 1))
(define p (ff:diff 3 o 'x))

(define q '(x -1 0 2 0 0 0 -2 0 1))

(define m1 '#2(( 0   0	 0   0	 0   0)
	       ( 3   4	-3  -3	-5   5)
	       ( 3  -5	 5   1	-1   0)
	       (-2   4	-1   2	-4  -2)
	       (-4  -3	-1   0	-1  -3)
	       (-3  -1	-4  -3	-1  -4)))

(define m2 '#2((0  0  0)
	       (1  2  3)
	       (4  5  6)))

(define m3 '#2((0  0  0	 0  0  0  0  0)
	       (2  0  7 11 10 12  5 11)
	       (3  6  3	 3  0  4  7  2)
	       (4  3  6	 4  1  6  2  3)
	       (2 11  8	 8  2  1  3 11)
	       (6 11  8	 6  2  6 10  9)
	       (5 11  7 10  0 11  6 12)
	       (3  3 12	 5  0 11  9 11)))


(define p1 '(x 1 2 1))
(define p2 '(x 1 3 3 1))
(define p3 '(x 1 4 6 4 1))
(define p4 '(x 1 -2 1))
(define p5 '(x 1 0 1))
(define p6 '(x -1 0 1))
(define p7 '(x 1 -3 -1 -3 1 -3 1))
(define p8 '(x 8 2 8 10 10 0 1 0 1))
(define p9 '(x 4 12 13 6 1))
(define p10 '(x 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1)) ; x^20+1
(define p11 '(x 5040 -432 10 1))	; GCL p234
(define p12 '(x 6 11 6 1))		; (x+1)(x+2)(x+3)
(define p13 '(x 7 0 8 7 1 8 0 1))	; (x+1)(x^2-x+1)(x^2+1)(x^2+7)
(define p14 '(x -1000 3650 10720 1415 -8198 -8235 -11288 -1894 9244
		4584 568 479 -46 1))
(define mv1 (list (sexp->var 'z) (list (sexp->var 'x) 0 2)  1 (list (sexp->var 'x) 0 0 1)))
(define mv2 (list (sexp->var 'z) (list (sexp->var 'x) 0 1)  1 1))
(define mv3 (list (sexp->var 'z) (list (sexp->var 'x) 10 7)  1 -8))
(define mv4 (list (sexp->var 'z) (list (sexp->var 'x) 0 2)  (list (sexp->var 'x) 0 0 1)))

; test data  (Amiga 1200 2Mb Chip RAM 4Mb Fast RAM 50MHz 68030 CPU No FPU WB 3.0)
; ff.scm 1  Jan 1995   28460 msec (8420 msec gc) 653689 cells work (62811 cells other)
; ff.scm 27 Jan 1995   27800 msec (7960 msec gc) 653681 cells work (62811 cells other)
;		       (using = instead of eqv?)
;		       25640	   7600		 573246		    62811
;		       (and speeding up ff:leading)
;		       24760	   7040		 572511		    62811
;		       (and using cadr in u:one? u:zero? etc)
;		       22260	   6380		 487313		    62811
;		       (and removing ff:unorm from ff:pp ff:cont)
;		       19340	   4960		 283401		   126751
;		       (and speeding up ff:unorm)
;		       15500	   3920		 241554		   100271
;		       (and using u:lose-lzeros)
;		       12160	   3020		 226016		    62811
;		       (and rewriting ff:unorm)
;		       11320	   2920		 222492		    45683
;		       (and rewriting ff:p*c)
;		       10000	   2740		 203486		    30979
;		       (and rewriting ff:p+p, ff:p-p)
;			9960	   2780		 197025		    30979
;		       (and alter ff:p+p, ff:p-p, ff:p/c)
;			9500	   2700		 191548		    19235
;			9080	   2280		 189661		    19235
;		       (and alter ff:p/p)
;   30 January 1995    17520	   4680		 377861		    19831
;		       (and went over to symmetric representation)
;   31 January 1995    17300	   4820		 367493		    19679
;		       (and rewrote ff:p/p, using new ff:p/p->qr)
;   1 February 1995    11220	   3160		 218731		    23131
;		       (and used ff:p/p->qr in ff:euclid-gcd)
;   7 February 1995    11560	   3340		 211306		    24035
;  17 February 1995    10940	   2900		 192896		    24035
;  23 February 1995	9880	   3320		 157728		    28139
;		       (using latest scm 4e2)
;   1 March    1995	9403	   2852		 152970		    28139
;		       (scm 4e2 with CCLO's, and added result sorting)
;
;  30 May      1995	9282	   1415		 181783		    26643

(define (ff:test)
  (test '(x 1)
	ff:euclid-gcd 23 j k)
  (test '(x 0 -1 0 0 1)
	ff:euclid-gcd 5 l m)
  (test '(x 1 0 1)
	ff:euclid-gcd 7 l m)
  (test '(x 4 3 1)
	ff:euclid-gcd 11 l m)
  (test '((x -4 1) (x 4 1) (x 5 0 1))
	ff:berlekamp 11 '(x 8 0 0 0 1))
  (test '((x 1 1) (x 3 5 1) (x 4 3 2 1) )
	ff:berlekamp 11 p7)
  (test '((x 3 1) (x -1 4 -5 1) (x 6 4 3 2 1))
	ff:berlekamp 13 p8)
  (test '(((x 1 1) 2))
	u:sff p1)
  (test '(((x -1 1) 2))
	u:sff p4)
  (test '(((x 2 3 1) 2))
	u:sff p9)
  (test '(((x 8 2 8 10 10 0 1 0 1) 1))
	u:sff p8)
  'done)

;;; ff:test2
;;;   4 February  1995	13260	    3360	  246229	       62499
;;;   1 March	  1995	13618	    4042	  208548	       63807
;;;			(scm 4e2 with CCLO's, and added result sorting)
;;;  30 May	  1995	10049	    1817	  164535	       43035

(define (ff:test2)
  (test '((x -1 -1 1) (x -1 1 1) (x 1 0 1 -1 1) (x 1 -1 1 0 1) (x 1 1 1 0 1) (x 1 0 1 1 1))
	 ff:berlekamp 3 p10)
  'done)

;;; ff:test3
;;;   4 February  1995	32100	    8020	  602668	       81739
;;;  23 February  1995	29840	    8440	  460900	       84943
;;;			(using latest scm 4e2)
;;;   1 March	  1995	29706	    8840	  454745	       84943
;;;			(scm 4e2 with CCLO's, and added result sorting)
;;;  30 May	  1995	27696	    4254	  480404	       69875

(define (ff:test3)
  (test '((x -5 0 1) (x 5 0 1) (x -1 1 2 -5 1) (x -1 5 -2 -1 1) (x -1 -5 -2 1 1) (x -1 -1 2 5 1))
	 ff:berlekamp 13 p10)
  'done)


;;; ff:test4
;;;  21 February  1995	74360 mSec (21780 in gc) 1391234 cells work, 59617 bytes other
;;;			29820	     8540	  539821	     24317
;;;			(after memoisation of hen:diophantine. 4 seconds on second run)
;;;			25980	     7560	  460937	     20637
;;;			(after memoization of hen:eea-lift etc)
;;;  22 February 1995	22040	     3660	  459685	     19653
;;;			(after removal of some un-needed error checking)
;;;  23 February 1995	26820	     8680	  466699	     21261
;;;   1 March	 1995	26462	     8463	  456924	     22793
;;;			(scm 4e2 with CCLO's, and added result sorting)
;;;  30 May	 1995	21597	     3257	  458526	     21721

(define (ff:test4)
  (test '((((x 1 1) (x 1 -1 1) (x 1 0 1) (x 7 0 1)) 1))
	u:factorz p13)
  'done)
